"""
QuarterBit
==========

Memory-efficient optimizer for PyTorch.

Usage:
    from quarterbit import Axiom
    optimizer = Axiom(model.parameters(), lr=1e-3)

Copyright (c) 2026 Clouthier Simulation Labs. All rights reserved.
"""

__version__ = "4.0.3"
__author__ = "Kyle Clouthier"

# Handle missing torch gracefully
try:
    from .torch.optim import Axiom
    __all__ = ['Axiom']
except ImportError as e:
    if 'torch' in str(e).lower():
        print("\n\033[93mQuarterBit requires PyTorch\033[0m")
        print("Install: \033[96mpip install torch\033[0m")
        print("Or with CUDA: \033[96mpip install torch --index-url https://download.pytorch.org/whl/cu121\033[0m\n")
    raise


def _check_license():
    """Check if license key is configured, show registration message if not."""
    import os
    import tempfile
    from pathlib import Path

    # Skip if already shown recently (once per week)
    cache_file = Path(tempfile.gettempdir()) / '.quarterbit_license_check'
    try:
        if cache_file.exists():
            import time
            if time.time() - cache_file.stat().st_mtime < 604800:  # 7 days
                return
    except:
        pass

    # Check for license key
    license_key = os.environ.get('QUARTERBIT_LICENSE_KEY', '')

    if not license_key:
        print("\n\033[96m" + "=" * 50 + "\033[0m")
        print("\033[1mWelcome to QuarterBit AXIOM!\033[0m")
        print("\033[96m" + "=" * 50 + "\033[0m")
        print("\nGet your \033[92mFREE license key\033[0m to activate:")
        print("  \033[96mhttps://quarterbit.dev/register\033[0m")
        print("\nThen set it:")
        print("  \033[93mexport QUARTERBIT_LICENSE_KEY=your-key-here\033[0m")
        print("\nFree tier: Personal use, research, <10 GPU-hrs/month")
        print("\033[96m" + "=" * 50 + "\033[0m\n")

        # Update cache
        try:
            cache_file.write_text('shown')
        except:
            pass


def _check_for_updates():
    """Check PyPI for newer version (runs in background thread)."""
    import os
    import threading

    # Skip if disabled
    if os.environ.get('QUARTERBIT_NO_UPDATE_CHECK', '').lower() in ('1', 'true', 'yes'):
        return

    def check():
        try:
            import json
            import urllib.request
            import tempfile
            import time
            from pathlib import Path

            # Cache file to avoid checking too often (once per day)
            cache_file = Path(tempfile.gettempdir()) / '.quarterbit_update_check'

            # Check if we already checked today
            if cache_file.exists():
                try:
                    cache_time = cache_file.stat().st_mtime
                    if time.time() - cache_time < 86400:  # 24 hours
                        return
                except:
                    pass

            # Query PyPI
            url = 'https://pypi.org/pypi/quarterbit/json'
            req = urllib.request.Request(url, headers={'User-Agent': 'quarterbit-update-check'})
            with urllib.request.urlopen(req, timeout=3) as response:
                data = json.loads(response.read().decode())
                latest = data['info']['version']

            # Update cache
            try:
                cache_file.write_text(latest)
            except:
                pass

            # Compare versions
            def parse_version(v):
                return tuple(int(x) for x in v.split('.')[:3])

            current = parse_version(__version__)
            latest_parsed = parse_version(latest)

            if latest_parsed > current:
                print(f"\n\033[93mQuarterBit {latest} available\033[0m (you have {__version__})")
                print(f"Upgrade: \033[96mpip install --upgrade quarterbit\033[0m\n")

        except:
            # Silently fail - don't interrupt user's workflow
            pass

    # Run in background thread
    thread = threading.Thread(target=check, daemon=True)
    thread.start()


# Run checks on import
_check_license()
_check_for_updates()



"""
QuarterBit PyTorch Backend
==========================

Usage:
    from quarterbit import Axiom
    optimizer = Axiom(model.parameters(), lr=1e-3)

Copyright (c) 2026 Clouthier Simulation Labs. All rights reserved.
"""

from .optim import Axiom

__all__ = ['Axiom']



"""
QuarterBit AXIOM Optimizer
==========================

Memory-efficient Adam variant with enhanced numerical precision.

Usage:
    from quarterbit import Axiom
    optimizer = Axiom(model.parameters(), lr=1e-3)

Copyright (c) 2026 Clouthier Simulation Labs. All rights reserved.
"""

import torch
from torch.optim import Optimizer
from typing import Optional, Callable, Iterable
import ctypes
import os


def _ptr(tensor):
    """Get ctypes pointer to tensor data."""
    if tensor is None:
        return None
    return ctypes.cast(tensor.data_ptr(), ctypes.POINTER(ctypes.c_float))


def _load_axiom_lib():
    """Load the AXIOM CUDA library."""
    if os.name == 'nt':
        lib_name = 'quarterbit_axiom.dll'
    else:
        lib_name = 'libquarterbit_axiom.so'

    search_paths = [
        os.path.join(os.path.dirname(__file__), '..', 'lib', lib_name),
        os.path.join(os.path.dirname(__file__), '..', '..', 'build', lib_name),
        os.path.join(os.path.dirname(__file__), '..', '..', '..', 'build', lib_name),
        os.path.join('C:', 'QuarterBit', 'build', lib_name),
        '/kaggle/working/build/' + lib_name,
    ]

    for path in search_paths:
        if os.path.exists(path):
            try:
                lib = ctypes.CDLL(path)

                lib.axiom_create.argtypes = [
                    ctypes.c_int,  # n
                    ctypes.c_int,  # total_steps
                    ctypes.c_int   # warmup_steps
                ]
                lib.axiom_create.restype = ctypes.c_void_p

                lib.axiom_step.argtypes = [
                    ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p,
                    ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float,
                    ctypes.c_int
                ]
                lib.axiom_step.restype = ctypes.c_int

                lib.axiom_set_schedule.argtypes = [
                    ctypes.c_void_p, ctypes.c_int, ctypes.c_int
                ]
                lib.axiom_set_schedule.restype = None

                lib.axiom_destroy.argtypes = [ctypes.c_void_p]
                lib.axiom_destroy.restype = None

                return lib
            except Exception:
                continue
    return None


class Axiom(Optimizer):
    """
    QuarterBit AXIOM Optimizer.

    World's most memory-efficient drift-free optimizer.
    90%+ memory savings vs Adam with full numerical precision.

    Args:
        params: Parameters to optimize
        lr: Learning rate (default: 1e-3)
        betas: Momentum coefficients (default: (0.9, 0.999))
        eps: Numerical stability term (default: 1e-8)
        weight_decay: Weight decay coefficient (default: 0)
        total_steps: Total training steps for LR scheduling (default: 0)
        warmup_steps: Warmup steps (default: 0)

    Example:
        >>> optimizer = Axiom(model.parameters(), lr=1e-3)
        >>> optimizer = Axiom(model.parameters(), lr=1e-3, total_steps=10000)
    """

    def __init__(
        self,
        params: Iterable[torch.nn.Parameter],
        lr: float = 1e-3,
        betas: tuple = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        total_steps: int = 0,
        warmup_steps: int = 0,
    ):
        if lr < 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if eps < 0.0:
            raise ValueError(f"Invalid epsilon: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta[0]: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta[1]: {betas[1]}")

        defaults = dict(
            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
            total_steps=total_steps, warmup_steps=warmup_steps
        )
        super().__init__(params, defaults)

        self._handles = {}
        self._step_count = 0
        self._total_steps = total_steps
        self._warmup_steps = warmup_steps

        self._lib = _load_axiom_lib()
        if self._lib is None:
            raise RuntimeError(
                "QuarterBit AXIOM CUDA library not found. "
                "Ensure CUDA is available and the library is installed."
            )

    def _get_handle(self, p):
        """Get or create handle for parameter."""
        if id(p) not in self._handles:
            if p.is_cuda and p.dtype == torch.float32:
                handle = self._lib.axiom_create(
                    ctypes.c_int(p.numel()),
                    ctypes.c_int(self._total_steps),
                    ctypes.c_int(self._warmup_steps)
                )
                self._handles[id(p)] = handle if handle else None
            else:
                self._handles[id(p)] = None
        return self._handles[id(p)]

    def set_schedule(self, total_steps: int, warmup_steps: int = 0):
        """
        Update training schedule.

        Args:
            total_steps: Total training steps
            warmup_steps: Warmup steps (default: 0)
        """
        self._total_steps = total_steps
        self._warmup_steps = warmup_steps

        for group in self.param_groups:
            group['total_steps'] = total_steps
            group['warmup_steps'] = warmup_steps

        for handle in self._handles.values():
            if handle is not None:
                self._lib.axiom_set_schedule(
                    handle,
                    ctypes.c_int(total_steps),
                    ctypes.c_int(warmup_steps)
                )

    @torch.no_grad()
    def step(self, closure: Optional[Callable] = None):
        """Perform a single optimization step."""
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        self._step_count += 1

        for group in self.param_groups:
            beta1, beta2 = group['betas']
            lr = group['lr']
            eps = group['eps']
            weight_decay = group['weight_decay']

            for p in group['params']:
                if p.grad is None:
                    continue

                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Axiom does not support sparse gradients")

                handle = self._get_handle(p)

                if handle is not None:
                    if weight_decay != 0:
                        grad = grad.add(p.data, alpha=weight_decay)

                    self._lib.axiom_step(
                        handle,
                        _ptr(p.data),
                        _ptr(grad),
                        ctypes.c_float(lr),
                        ctypes.c_float(beta1),
                        ctypes.c_float(beta2),
                        ctypes.c_float(eps),
                        ctypes.c_int(self._step_count)
                    )
                else:
                    raise RuntimeError(
                        f"Parameter not on CUDA or not float32. "
                        f"Got device={p.device}, dtype={p.dtype}"
                    )

        return loss

    def __del__(self):
        """Clean up handles."""
        if hasattr(self, '_handles') and hasattr(self, '_lib') and self._lib is not None:
            for handle in self._handles.values():
                if handle is not None:
                    try:
                        self._lib.axiom_destroy(handle)
                    except Exception:
                        pass



"""
QuarterBit Build Script
=======================

Compiles Python to binary extensions using Cython for IP protection.

Usage:
    python setup.py build_ext --inplace   # Build locally
    python setup.py bdist_wheel           # Build wheel for PyPI

Copyright (c) 2026 Clouthier Simulation Labs. All rights reserved.
"""

import os
import sys
import platform
from setuptools import setup, find_packages
from setuptools.extension import Extension

# Check for Cython
try:
    from Cython.Build import cythonize
    from Cython.Distutils import build_ext
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False
    print("Cython not found - building without compilation")

# Version
VERSION = "4.0.3"

# Modules to compile (protect these)
CYTHON_MODULES = [
    "quarterbit/torch/optim.py",
]

# Platform-specific library files
if platform.system() == "Windows":
    LIB_FILES = ["lib/*.dll"]
else:
    LIB_FILES = ["lib/*.so"]

# Create extensions if Cython available
extensions = []
if USE_CYTHON:
    for module_path in CYTHON_MODULES:
        if os.path.exists(module_path):
            module_name = module_path.replace("/", ".").replace("\\", ".").replace(".py", "")
            extensions.append(
                Extension(
                    module_name,
                    [module_path],
                    define_macros=[("CYTHON_TRACE", "0")],
                )
            )

# Cython compiler directives
cython_directives = {
    "language_level": "3",
    "boundscheck": False,
    "wraparound": False,
    "cdivision": True,
    "embedsignature": False,  # Hide API signatures
}

# Setup configuration
setup_kwargs = dict(
    name="quarterbit",
    version=VERSION,
    description="Memory-efficient optimizer for PyTorch",
    long_description=open("README.md").read() if os.path.exists("README.md") else "",
    long_description_content_type="text/markdown",
    author="Kyle Clouthier",
    author_email="info@clouthiersimulation.com",
    url="https://github.com/clouthiersimulation/quarterbit",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "torch>=1.8",
        "numpy>=1.20",
    ],
    package_data={
        "quarterbit": LIB_FILES,
    },
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="optimizer pytorch deep-learning machine-learning",
)

# Add Cython compilation if available
if USE_CYTHON and extensions:
    setup_kwargs["ext_modules"] = cythonize(
        extensions,
        compiler_directives=cython_directives,
        annotate=False,
    )
    setup_kwargs["cmdclass"] = {"build_ext": build_ext}

setup(**setup_kwargs)
